# cython: language_level=3
# cython: boundscheck=False
# cython: cdivision=True
# cython: wraparound=False

import numpy as np
from ..retinavision.retina import Retina as old_retina
from ..retinavision.rf_generation import rf_ozimek
from cython.parallel import parallel
from cython.parallel import prange

cimport cython
cimport numpy as cnp
from google.colab.patches import cv2_imshow
import cv2

from src.software_retina.utils cimport multiply_and_sum2d
from src.software_retina.utils cimport multiply_and_sum3d
from src.software_retina.utils cimport pad_coloured
from src.software_retina.utils cimport pad_grayscaled

# Original code provided by Piotr Ozimek

cdef class Retina:

    cdef int N, width
    cdef cnp.float64_t[:, ::1] node_attributes
    cdef cnp.int32_t[:, :, ::1] coefficients
    cdef cnp.float64_t[::1] grayscale_intensity
    cdef cnp.float64_t[:, ::1] colour_intensity

    def __init__(self, input_node_attributes, input_coefficients):

        self.node_attributes = input_node_attributes
        self.coefficients = input_coefficients
        self.N = len(input_node_attributes)
        self.width = 2*int(np.abs(input_node_attributes[:, :2]).max() +
                           input_node_attributes[:, 6].max()/2.0)
        self.grayscale_intensity = np.zeros((1))
        self.colour_intensity = np.zeros((1, 1))

    def load_node_attributes(self, input):
        if isinstance(input, np.ndarray):
            if not (input.ndim == 2 and input.shape[1] == 7):
                raise ValueError('Must be a 2 dimensional array with each row'
                                 ' b having 7 columns of node attributes')

            else:
                self.node_attributes = input
                self.N = len(self.node_attributes)
                self.width = 2*int(
                    np.abs(self.node_attributes[:, :2]).max() +
                    np.asarray(self.node_attributes[:, 6]).max()/2.0)

    def load_coefficients(self, input):
        if isinstance(input, np.ndarray):
            if not input.ndim == 3:
                raise ValueError('Must be 3 dimensional array')

            else:
                self.coefficients = input
        else:
            raise TypeError('This function only accepts numpy array')

    cpdef cnp.ndarray[cnp.float64_t, ndim=1] sample_grayscale(
            self,
            cnp.ndarray[cnp.uint8_t, ndim=2] image,
            (int, int) fixation):

        cdef int fixation_y = fixation[0]
        cdef int fixation_x = fixation[1]
        cdef int p, y1, y2, x1, x2
        cdef cnp.int32_t[:, ::1] pic
        cdef cnp.ndarray[cnp.float64_t, ndim=1] X
        cdef cnp.ndarray[cnp.float64_t, ndim=1] Y
        cdef cnp.ndarray[cnp.float64_t, ndim=1] V
        cdef Py_ssize_t i
        cdef float w

        p = self.width
        pic = pad_grayscaled(image, p)
        X = np.asarray(self.node_attributes[:, 0]) + fixation_x + p
        Y = np.asarray(self.node_attributes[:, 1]) + fixation_y + p
        V = np.empty((self.N))

        with nogil, parallel():
            for i in prange(self.N):
                w = self.node_attributes[i, 6]
                y1 = int(Y[i] - w/2+0.5)
                y2 = int(Y[i] + w/2+0.5)
                x1 = int(X[i] - w/2+0.5)
                x2 = int(X[i] + w/2+0.5)
                V[i] = multiply_and_sum2d(pic[y1:y2, x1:x2],
                                          self.coefficients[i, :, :])

        self.grayscale_intensity = V

        return V

    cpdef cnp.ndarray[cnp.float64_t, ndim=2] sample_colour(
            self,
            cnp.ndarray[cnp.uint8_t, ndim=3] image,
            (int, int) fixation):

        cdef int fixation_y = fixation[0]
        cdef int fixation_x = fixation[1]
        cdef int p, y1, y2, x1, x2
        cdef cnp.int32_t[:, :, ::1] pic
        cdef cnp.ndarray[cnp.float64_t, ndim=1] X
        cdef cnp.ndarray[cnp.float64_t, ndim=1] Y
        cdef Py_ssize_t i
        cdef float w
        cdef cnp.float64_t[:, ::1] V = np.empty((self.N, 3))
        cdef cnp.float64_t[::1] sum3d_return = np.empty((3))

        p = self.width
        pic = pad_coloured(image, p)
        X = np.asarray(self.node_attributes[:, 0]) + fixation_x + p
        Y = np.asarray(self.node_attributes[:, 1]) + fixation_y + p

        with nogil, parallel():
            for i in prange(self.N):
                w = self.node_attributes[i, 6]
                y1 = int(Y[i] - w/2+0.5)
                y2 = int(Y[i] + w/2+0.5)
                x1 = int(X[i] - w/2+0.5)
                x2 = int(X[i] + w/2+0.5)
                V[i] = multiply_and_sum3d(pic[y1:y2, x1:x2, :],
                                          self.coefficients[i, :, :],
                                          sum3d_return)

        self.colour_intensity = V

        return np.asarray(V)

    # wrapper to compare foveated images with ozimek's old code
    def compare_foveated_images(self, input_tessellation, image):
        rf_loc, rf_coeff, fov_5 = rf_ozimek(input_tessellation, kernel_ratio = 3, sigma_base = 0.5, sigma_power = 1, mean_rf = 1)
        ozimek_retina = old_retina()
        ozimek_retina.loadLoc(rf_loc)
        ozimek_retina.loadCoeff(rf_coeff)
        image_gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        y, x = image_gray.shape
        y = int(y/2)
        x = int(x/2)
        ozimek_retina.sample(image_gray, (y, x))
        ozimek_backproject_5k = ozimek_retina.backproject_tight_last()
        ozimek_retina.sample(image_gray, (y, x))
        ozimek_retina._V = self.sample_grayscale(image_gray, (y, x))
        our_backproject_5k = ozimek_retina.backproject_tight_last()
        cv2_imshow(ozimek_backproject_5k)
        cv2_imshow(our_backproject_5k)

        difference = ozimek_backproject_5k - our_backproject_5k
        difference = difference*1000
        cv2_imshow(difference) 




import time
import sys

import numpy as np
import pynanoflann
from scipy.spatial.distance import cdist

from .utils import normalize, cartesian_to_polar, polar_to_cartesian

# Original code provided by George Killick


# set seed for evaluation
rng = np.random.default_rng(seed=12345)


class SelfSimilarNeuralNetwork:

    def __init__(self, node_count, foveal_region_size,
                 nearest_neighbour_method='auto'):

        self.node_count = node_count
        self.foveal_region_size = foveal_region_size
        self.weights = SelfSimilarNeuralNetwork.__init_weights(node_count)
        self.nearest_neighbour_method = nearest_neighbour_method
        self.nanoflann = pynanoflann.KDTree(n_neighbors=1, metric='L2',
                                            radius=1)

    def fit(self, num_iters=20000, initial_learning_rate=0.1,
            final_learning_rate=0.0005, verbose=True):

        learning_rate = SelfSimilarNeuralNetwork.__alpha_schedule(
            initial_learning_rate, final_learning_rate,
            num_iters, num_iters//4)
        start = time.time()
        get_neighbours = self.__select_nearest_neighbour_method(
            self.nearest_neighbour_method)

        for i in range(num_iters):
            alpha = learning_rate[i]

            input_vectors = np.copy(self.weights)
            input_vectors = cartesian_to_polar(input_vectors)

            d = np.exp((2*rng.uniform() - 1)*np.log(8))

            input_vectors[:, 1] *= d
            input_vectors = polar_to_cartesian(input_vectors)

            delta_theta = 2*rng.uniform()*np.pi
            delta_rho = rng.uniform() * self.foveal_region_size

            input_vectors[:, 0] += np.cos(delta_theta)*delta_rho
            input_vectors[:, 1] += np.sin(delta_theta)*delta_rho
            input_vectors = cartesian_to_polar(input_vectors)
            input_vectors[:, 0] += 2*rng.uniform()*np.pi

            cull = np.where(input_vectors[:, 1] <= 1)[0]

            input_vectors = polar_to_cartesian(input_vectors)
            input_vectors = input_vectors[cull]

            index = get_neighbours(input_vectors, self.weights)
            self.weights[index] -= ((self.weights[index] - input_vectors)
                                    * alpha)
            if (verbose):
                sys.stdout.write('\r' + str(i + 1) + "/" + str(num_iters))

        normalize(self.weights)
        if(verbose):
            print("\nFinished.")
            print("Time taken: " + str(time.time()-start))

    def set_weights(self, X):

        self.node_count = X.shape[0]
        self.weights = X

    def __select_nearest_neighbour_method(self, nearest_neighbour_method):

        if (nearest_neighbour_method == 'brute_force'):
            print("Using bruteforce.")
            return self.__brute_force_neighbours

        elif(nearest_neighbour_method == 'nanoflann'):
            print("Using nanoflann method with " + str(rng.bit_generator))
            return self.__pynanoflann_neighbours

        elif(nearest_neighbour_method == 'nanoflann_multi_jobs'):
            print("Using nanoflann method with 8 jobs " +
                  str(rng.bit_generator))
            return self.__pynanoflann_multi_neighbours_8

        elif(nearest_neighbour_method == 'auto'):
            if(self.node_count <= 256):
                print("Using bruteforce.")
                return self.__brute_force_neighbours

            elif(self.node_count <= 14999):
                print("Using nanoflann method.")
                return self.__pynanoflann_neighbours

            else:
                print("Using nanoflann method with 8 jobs.")
                return self.__pynanoflann_multi_neighbours_8

        else:
            print("Unknown nearest_neighbour_method, using nanoflann.")

    def __brute_force_neighbours(self, update_vectors, network_weight):

        dists = cdist(update_vectors, network_weight)
        indices = np.argmin(dists, axis=1)

        return indices

    def __pynanoflann_neighbours(self, update_vectors, network_weight):

        self.nanoflann.fit(network_weight)
        distances, indices = self.nanoflann.kneighbors(update_vectors)

        return indices.flatten()

    def __pynanoflann_multi_neighbours_8(self, update_vectors, network_weight):

        self.nanoflann.fit(network_weight)
        distances, indices = self.nanoflann.kneighbors(update_vectors,
                                                       n_jobs=8)

        return indices.flatten()

    @staticmethod
    def __init_weights(node_count):

        r = rng.uniform(1, 0, node_count)
        th = 2*np.pi*rng.uniform(1, 0, node_count)

        return polar_to_cartesian(np.array([th, r]).T)

    @staticmethod
    def __alpha_schedule(initial_learning_rate, final_learning_rate,
                         num_iters, split):

        """ Creates a learning rate schedule for the SelfSimilarNeuralNetwork.
            Constant learning rate for first "split" of iterations
            and then linearly annealing for the remainder.

            Parameters
            ----------
            initial_learning_rate: starting learning rate.
            final_learning_rate: final learning rate.
            num_iters: number of iterations, equivalent to number
            of iterations to train the network for.
            split: decides when to start annealing, typical after 25%
            oftotal iterations.

            Return: Numpy array for learning rate; length num_iters.

        """

        static = split
        decay = num_iters - static
        static_lr = np.linspace(initial_learning_rate,
                                initial_learning_rate,
                                static)
        decay_lr = np.linspace(initial_learning_rate,
                               final_learning_rate,
                               decay)

        return np.concatenate((static_lr, decay_lr))



# cython: language_level=3
# cython: boundscheck=False
# cython: cdivision=True
# cython: wraparound=False

import numpy as np
cimport numpy as cnp

from libc.math cimport sqrt, exp, pi


cpdef inline cnp.int32_t[:, ::1] pad_grayscaled(
        cnp.ndarray[cnp.uint8_t, ndim=2] img,
        int padding):

    cdef cnp.int32_t[:, ::1] image_mem_view = img.astype(dtype=np.int32)
    cdef int first_dimension = img.shape[0]
    cdef int first_accumulator = 0
    cdef int second_dimension = img.shape[1]
    cdef int second_accumulator = 0
    cdef int padding_twice = 2*padding
    cdef cnp.int32_t[:, ::1] out
    cdef Py_ssize_t i
    cdef int img_dimension = img.ndim

    for i in range(img_dimension):
        if i == 1:
            first_accumulator += first_dimension + padding_twice

        else:
            second_accumulator += second_dimension + padding_twice

    out = np.zeros((first_accumulator, second_accumulator), dtype=np.int32)
    out[padding:-padding, padding:-padding] = image_mem_view

    return out

cpdef inline cnp.int32_t[:, :, ::1] pad_coloured(
        cnp.ndarray[cnp.uint8_t, ndim=3] img,
        int padding):

    cdef cnp.int32_t[:, :, ::1] image_mem_view = img.astype(dtype=np.int32)
    cdef int first_dimension = img.shape[0]
    cdef int first_accumulator = 0
    cdef int second_dimension = img.shape[1]
    cdef int second_accumulator = 0
    cdef int third_dimension = img.shape[2]
    cdef int padding_twice = 2*padding
    cdef cnp.int32_t[:, :, ::1] out
    cdef Py_ssize_t i
    cdef int img_dimension = img.ndim

    for i in range(img_dimension):
        if i == 1:
            first_accumulator += first_dimension + padding_twice

        if i == 2:
            second_accumulator += second_dimension + padding_twice

    out = np.zeros((first_accumulator, second_accumulator, third_dimension),
                   dtype=np.int32)
    out[padding:-padding, padding:-padding, :] = image_mem_view

    return out

cdef inline double multiply_and_sum2d(
        cnp.int32_t[:, ::1] image_extract,
        cnp.int32_t[:, ::1] coeff_mem_view) nogil:

    cdef size_t i, first_dimension, j, second_dimension
    cdef double total = 0
    cdef signed long long x
    cdef signed long long y
    first_dimension = image_extract.shape[0]
    second_dimension = image_extract.shape[1]

    for i in range(first_dimension):
        for j in range(second_dimension):
            x = image_extract[i, j]
            y = coeff_mem_view[i, j]
            total += x*y

    return total/100000000


cdef inline cnp.float64_t[::1] multiply_and_sum3d(
        cnp.int32_t[:, :, ::1] image_extract,
        cnp.int32_t[:, ::1] coeff_mem_view,
        cnp.float64_t[::1] sum3d_return) nogil:

    cdef size_t i, first_dimension, j, second_dimension, k, third_dimension
    cdef double total = 0
    first_dimension = image_extract.shape[0]
    second_dimension = image_extract.shape[1]
    third_dimension = image_extract.shape[2]
    cdef signed long long x
    cdef signed long long y
    cdef signed long long column_0 = 0
    cdef signed long long column_1 = 0
    cdef signed long long column_2 = 0
    cdef double total_column_0 = 0
    cdef double total_column_1 = 0
    cdef double total_column_2 = 0

    for i in range(first_dimension):
        for j in range(second_dimension):
            for k in range(third_dimension):
                if k == 0:
                    x = image_extract[i, j, k]
                    y = coeff_mem_view[i, j]
                    total_column_0 += x*y

                if k == 1:
                    x = image_extract[i, j, k]
                    y = coeff_mem_view[i, j]
                    total_column_1 += x*y

                if k == 2:
                    x = image_extract[i, j, k]
                    y = coeff_mem_view[i, j]
                    total_column_2 += x*y

    sum3d_return[0] = total_column_0/100000000
    sum3d_return[1] = total_column_1/100000000
    sum3d_return[2] = total_column_2/100000000

    return sum3d_return

cpdef inline cnp.float64_t gauss(
        cnp.float64_t sigma,
        cnp.float64_t x,
        cnp.float64_t y,
        int mean):

    cdef cnp.float64_t d

    d = sqrt(x*x + y*y)

    return exp(-(d-mean)**2 / (2*sigma**2)) / sqrt(2*pi*sigma**2)


cpdef inline cnp.ndarray[cnp.float64_t, ndim=2] gausskernel(
        cnp.int_t width,
        cnp.ndarray[cnp.float64_t, ndim=1] loc,
        cnp.float64_t sigma):

    cdef cnp.ndarray[cnp.float64_t, ndim=2] k
    cdef double w, shift, dx, dy
    cdef int x, y

    w = float(width)
    k = np.zeros((width, width))
    shift = (w - 1) / 2.0

    dx = loc[0] - int(loc[0])
    dy = loc[1] - int(loc[1])

    for x in range(width):
        for y in range(width):
            k[y, x] = gauss(sigma, (x-shift) - dx, (y-shift) - dy, 0)

    return k



import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial import KDTree, Delaunay

# Original code provided by George Killick


def cartesian_to_polar(coordinates):

    x = coordinates[:, 0]
    y = coordinates[:, 1]
    rho = np.sqrt(x**2 + y**2)
    theta = np.arctan2(y, x)

    return np.stack([theta, rho], 1)


def polar_to_cartesian(coordinates):

    theta = coordinates[:, 0]
    rho = coordinates[:, 1]
    x = rho*np.cos(theta)
    y = rho*np.sin(theta)

    return np.stack([x, y], 1)


def display_tessellation(points, figsize=(15, 15), s=1):

    plt.figure(figsize=figsize)
    plt.scatter(points[:, 0], points[:, 1], s=s, c='black')
    plt.show()

    return


def display_stats(points, figsize=(15, 8), k=7):

    nbs = KDTree(points)

    distance, index = nbs.query(points, k=k)

    eccentricity = np.linalg.norm(points, axis=1)
    distance = 1/np.mean(distance[:, 1:], axis=1)

    distance = distance[np.argsort(eccentricity)]
    eccentricity = np.sort(eccentricity)

    plt.figure(figsize=figsize)
    plt.plot(eccentricity, distance)
    plt.xlabel("Eccentricity")
    plt.ylabel("1/mean distance to neighbours")
    plt.show()

    return


def normalize(points):

    points = cartesian_to_polar(points)
    points[:, 1] /= np.max(points[:, 1])
    points = polar_to_cartesian(points)

    return points


def randomize(points, alpha=0.25):

    x = np.copy(points)
    nbs = KDTree(x)

    distance, index = nbs.query(x, k=5)
    distance = np.mean(distance[:, 1:], axis=1)

    delta_rho = distance*np.sqrt(np.random.uniform())
    delta_theta = 2*np.random.uniform(size=distance.shape[0])*np.pi
    delta_x = delta_rho*np.cos(delta_theta)
    delta_y = delta_rho*np.sin(delta_theta)

    x[:, 0] += (delta_x*alpha)
    x[:, 1] += (delta_y*alpha)

    return normalize(x)


def point_generation(points, mode='sierpinski', concatenate=True):

    # Generates points using delaunay triangulation
    # and barycentre or sierpinski methods for generating
    # points.

    points = points
    triangulation = Delaunay(points)
    d_tri = points[triangulation.simplices]

    if(mode == 'barycentre'):
        new_points = np.sum(d_tri, axis=1)/3

    elif(mode == 'sierpinski'):

        a = (d_tri[:, 0] + d_tri[:, 1])/2
        b = (d_tri[:, 1] + d_tri[:, 2])/2
        c = (d_tri[:, 2] + d_tri[:, 0])/2

        new_points = np.concatenate((a, b, c), axis=0)
        new_points = np.unique(new_points, axis=0)

    else:
        print("Unknown mode; Choose from 'barycentre' or 'sierpinski'.")
        return None

    if(concatenate):
        new_points = np.concatenate((points, new_points), axis=0)

    return normalize(new_points)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 26 19:24:06 2018

Utility functions used throughout the project

TODO: create a separate tessellation utils file

@author: Piotr Ozimek
"""
import sys
py = sys.version_info.major

import matplotlib.pyplot as plt
import cv2
import numpy as np


if py == 2: import cPickle as pickle
elif py == 3: 
    import torch
    import pickle


"""Matrix manipulation functions""" 
#Pad an image (with 0s or NaNs) from all sides
def pad(img, padding, nans=False):
    size = ()
    for i in range(len(img.shape)):
        if i != 2: size += (img.shape[i] + 2*padding,)
        else: size += (img.shape[i],)

    out = np.zeros(size, dtype = img.dtype)
    
    if nans: out = np.full(size, np.nan)
    out[padding:-padding, padding:-padding] = img
    
    return out

#Project the source image onto the target image at the given location
def project(source, target, location, v=False):
    sh, sw = source.shape[:2]
    th, tw = target.shape[:2]
    
    #target frame
    y1 = max(0, ir(location[0] - sh/2.0))
    y2 = min(th, ir(location[0] + sh/2.0))
    x1 = max(0, ir(location[1] - sw/2.0))
    x2 = min(tw, ir(location[1] + sw/2.0))
    
    #source frame
    s_y1 = - ir(min(0, location[0] - sh/2.0 + 0.5))
    s_y2 = s_y1 + (y2 - y1)
    s_x1 = - ir(min(0, location[1] - sw/2.0 + 0.5))
    s_x2 = s_x1 + (x2 - x1)
    
    try: target[y1:y2, x1:x2] += source[s_y1:s_y2, s_x1:s_x2]
    except Exception as E:
        print(y1, y2, x1, x2)
        print(s_y1, s_y2, s_x1, s_x2)
        print(source.shape)
        print(target.shape)
        print(location)
        raise E
    
    if v:
        print(y1, y2, x1, x2)
        print(s_y1, s_y2, s_x1, s_x2)
        print(source.shape)
        print(target.shape)
        print(location)
    
    return target


""" Convenience functions""" 
def bgr2rgb(im):
    return np.dstack((im[:,:,2], im[:,:,1], im[:,:,0]))

#OH BOY PYTHON 3 SURELY HURTS
def normal_round(n):
    if n - np.floor(np.abs(n)) < 0.5:
        return np.floor(n)
    return np.ceil(n)

#i = int, r = round.
def ir(val):
    return int(normal_round(val))
       
#Unused 
def normalize(V):
    return V *(255/V.max())

#Get GPU memory stats
def memshow():
    total = torch.cuda.get_device_properties(0).total_memory
    a = torch.cuda.memory_allocated()
    c = torch.cuda.memory_cached()
    r = total-torch.cuda.memory_allocated()
    print("Allocated:", a)
    print("Cached:", c)
    print("Remaining:", r, "(", r/total,")")

#Gaussian receptive field calculation
def d_gauss(self, sigma, d, normalize=True):
    g = np.exp(-(d)**2/(2*sigma**2))/np.sqrt(2*np.pi*sigma**2)
    if normalize: return g/np.sum(g)
    else: return g


""" Pickling functions. Pain ahead. 
Py36 Pickle is not compatible with Py2 cPickle, so latin1 encoding is needed."""
def loadPickle(path):
    with open(path, 'rb') as handle:
        if py == 3: return pickle.load(handle, encoding='latin1')
        return pickle.load(handle)
    
def writePickle(path, obj):
    with open(path, 'wb') as handle:
        pickle.dump(obj, handle)

#Use this function to open pickled files in py2 that wont open otherwise
#The associated bug usually mentions a missing module.        
def loadPickleNonbin(path):
    with open(path, 'r') as handle:
        if py == 3: return pickle.load(handle, encoding='latin1')
        return pickle.load(handle)        

"""Camera and visualisation functions""" 
def camopen():
    cap = 0
    camid = 0
    cap = cv2.VideoCapture(camid)
    cap.set(cv2.CAP_PROP_FRAME_HEIGHT, 720)
    cap.set(cv2.CAP_PROP_FRAME_WIDTH, 1080)
    
    while not cap.isOpened():
        print(str(camid) + ' failed, retrying\n')
        cv2.VideoCapture(camid).release()
        cap = cv2.VideoCapture(camid)
        camid += 1  
    return cap

#Run this if cam stops working. Does not work in py3 (cam never closes?)
def camclose(cap):
    cap.release()
    cv2.destroyAllWindows()

#Take a pic with the webcam, return RGB!
def snap():
    cap = camopen()
    ret, img = cap.read()
    camclose(cap)
    return bgr2rgb(img)

def picshow(pic, size=(10,10), cmap = 'gray'):    
    plt.figure(figsize=size)
    plt.axis('off')
    plt.imshow(pic, interpolation='none', cmap=cmap)
    plt.show()
    
def imshow(pic, size=(10,10)):    
    plt.figure(figsize=size)
    plt.axis('off')
    plt.imshow(pic, interpolation='none')
    plt.show()
    
def scatter(X,Y, size=(10,10)):    
    plt.figure(figsize=size)
    plt.scatter(X, Y)
    plt.show()
    
"""Tessellation specific functions"""
def tessplot(T, s = (10,10), c = None, cmap=None, size=1, axis='on', marker='o'):    
    plt.figure(figsize = s)
    if axis=='off': plt.axis('off')
    plt.scatter(T[:,0], T[:,1], s = size, c = c, cmap=cmap, marker=marker)
    plt.show()
    
    
"""GPU accelerated cdist. y = rows, x = columns. Demons ahead."""
def cdist_torch(x, y):
    xs, ys = x.shape[0], y.shape[0]
    dims = x.shape[1] if len(x.shape) == 2 else 0
    if dims: assert(x.shape[1] == y.shape[1])
    
    #Pre-compute required GPU memory. Assuming float64
    batch_process = False
    x_mem, y_mem, xd_mem, dist_mem = 8, 8, 8, 8
    for i in x.shape: x_mem *= i
    for i in y.shape: y_mem *= i
    for i in ys, xs: dist_mem *= i
    if dims:
        for i in ys, xs, dims: xd_mem *= i
    else:
        xd_mem = dist_mem
    mem = dist_mem + xd_mem #peak memory usage
    
    total = torch.cuda.get_device_properties(0).total_memory - 2000 
    if mem > total: 
        batch_process = True
        assert(total - (x_mem + y_mem) > 0), "Inputs too large to fit on GPU"
        print("cdist_torch: Insufficient GPU memory, computing in batches...")
        xd_space = total - (x_mem + y_mem)
        #print("Space for xd = ", xd_space,", xd size = ", xd_mem)
        
        
    #Start processing cdist
    Y = torch.from_numpy(y).to("cuda")
    X = torch.from_numpy(x).to("cuda")
    
    if batch_process:
        xd_bsize = xd_space // (8 * xs * dims) #size of a batch in rows or columns
        xd_batches = int(np.ceil(xd_mem/xd_space)) #number of batches
        Xd = torch.zeros(ys, xs, dims, device="cpu")
        
        for B in range(xd_batches):
            xd1 = B * xd_bsize
            xd2 = min((B+1) * xd_bsize, ys)
            xd_size = xd2-xd1
            
            if dims: Xdbatch = torch.zeros(xd_size, xs, dims, device="cuda")
            else: Xdbatch = torch.zeros(xd_size, xs, device="cuda")
              
            for i in range(xd1,xd2):
                Xdbatch[i-xd1] = X - Y[i]
            
            Xd[xd1:xd2] = Xdbatch
            del(Xdbatch)
        del X, Y
        dist_bsize = total // ((8 * xs)+ (8 * xs * dims))
        #gotta hold a row of dist and Xd simultaneously
        dist_batches = np.ceil((dist_mem+xd_mem)/total)
        dist = torch.zeros(ys, xs, device="cpu")
        
        for B in range(int(dist_batches)):
            dist1 = B * dist_bsize
            dist2 = min((B+1) * dist_bsize, ys)
            Xd_batch = Xd[dist1:dist2].to("cuda")
            
            if dims: d = torch.norm(Xd_batch,2, dims)
            else: d = np.abs(Xd) #Can't be bothered doing one more dims check
            
            dist[dist1:dist2] = d
            del d, Xd_batch
        out = dist
            
    else: #nice and easy, all in one gulp
        if dims: Xd = torch.zeros(ys,xs,dims,device="cuda")
        else: Xd = torch.zeros(ys,xs,device="cuda")
            
        for i in range(ys):
            Xd[i] = X - Y[i]       
        del X, Y
        
        if dims: dist = torch.norm(Xd,2, dims) #OOM here
        else: dist = np.abs(Xd)
        del Xd
        out = dist.to("cpu")
    
    torch.cuda.empty_cache()
    return out


# cython: language_level=3
# cython: boundscheck=False
# cython: cdivision=True
# cython: wraparound=False

import numpy as np
cimport numpy as cnp

from scipy.spatial import distance

from src.software_retina.utils cimport gauss
from src.software_retina.utils cimport gausskernel

# Original code provided by Piotr Ozimek


def node_attribute_kernel_generation(tessellation, kernel_ratio, sigma_base,
                                     sigma_power, min_rf, min_kernel=3):

    """ Generates node attributes and kernels for each node
        Parameters
        ----------
        tessellation: raw node locations [x,y] array
        kernel_ratio: the ratio of kernel to local node density (dist_5)
        sigma_base: the base sigma, or global sigma scaling factor (lambda)
        sigma_power: the power term applied to sigma scaling with eccentricity
        min_rf: sets the min distance between the 20 most central nodes
        min_kernel: imposes a minimum kernel width for the receptive fields
                    (default value: 3)

        Return
        rf_node_attributes - a num_of_nodes x 7 array that describes
                             each node as follows:
        [x, y, d, angle (radians), dist_5, rf_sigma, rf_width]
        rf_coefficients - an padded array of gaussian receptive field kernels

    """

    rf_node_attributes = np.zeros((len(tessellation), 7))
    rf_coefficients_unpadded = np.ndarray((1, len(tessellation)),
                                          dtype='object')

    row_length = 0
    column_length = 0
    output_list = []
    neighbourhood = 6

    print("rf generation - might take a while...")

    length = len(tessellation)
    chunk = 5000
    num = length/chunk
    if length % chunk != 0:
        num += 1

    dist_5 = np.zeros(length, dtype='float64')
    print(str(chunk) + " nodes in one chunk.")

    for i in range(int(num)):
        print("Processing chunk " + str(i))
        d = distance.cdist(tessellation[i*chunk:(i + 1)*chunk],
                           tessellation)
        s = np.sort(d)
        dist_5[i*chunk:(i + 1)*chunk] = np.mean(s[:, 1:neighbourhood], 1)

    fov_dist_5 = np.min(dist_5[:20])
    rf_node_attributes[:, :2] = tessellation*(1/fov_dist_5)*min_rf
    dist_5 = dist_5*(1/fov_dist_5)*min_rf
    rf_node_attributes[:, 3] = np.arctan2(rf_node_attributes[:, 1],
                                          rf_node_attributes[:, 0])
    rf_node_attributes[:, 4] = dist_5

    print("All chunks done.")

    correction = 0
    rf_node_attributes[:, 5] = sigma_base*(dist_5+correction)**sigma_power

    for i in range(len(tessellation)):
        kernel_width = max(min_kernel, int(np.ceil(kernel_ratio *
                                           rf_node_attributes[i, 4])))
        rf_node_attributes[i, 6] = kernel_width
        cx, cy = return_offsets(rf_node_attributes[i, 0],
                                rf_node_attributes[i, 1],
                                kernel_width)

        rx = rf_node_attributes[i][0] - cx
        ry = rf_node_attributes[i][1] - cy
        loc = np.array([rx, ry])
        rf_node_attributes[i, 2] = np.linalg.norm(rf_node_attributes[i, :2])

        rf_node_attributes[i, 0] = cx
        rf_node_attributes[i, 1] = cy
        rf_coefficients_unpadded[0, i] = gausskernel(kernel_width, loc,
                                                     rf_node_attributes[i, 5])
        rf_coefficients_unpadded[0, i] /= np.sum(
                                            rf_coefficients_unpadded[0, i])

        if rf_coefficients_unpadded[0, i].shape > (row_length, column_length):
            row_length, column_length = rf_coefficients_unpadded[0, i].shape

    print("Padding kernels now")

    rf_coefficients = np.zeros((len(tessellation), row_length, row_length))

    for i in range(len(tessellation)):
        x, y = rf_coefficients_unpadded[0][i].shape
        rf_coefficients[i, :x, :y] = rf_coefficients_unpadded[0][i]

    rf_coefficients = (rf_coefficients*100000000).astype(np.int32)

    return rf_node_attributes, rf_coefficients


def return_offsets(x, y, kernel_width):
    kernel_width = int(kernel_width)

    if kernel_width % 2 != 0:
        cx = round(x)
        cy = round(y)

    else:
        cx = round(x) + np.sign(x-round(x))*0.5
        cy = round(y) + np.sign(y-round(y))*0.5

    return cx, cy



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Sat Apr 14 19:50:12 2018
PhD retina. Object model and code cleanup.
@author: Piotr Ozimek
"""

import numpy as np
from .utils import pad, loadPickle, project, loadPickleNonbin

#TODO: do something about coeff being (1,X) in shape instead of (X)...
#TODO: a check in utils for whether CUDA is installed, print warnings if it isnt
class Retina:
    def __init__(self, gpu=False):
        self.loc = 0
        self.N = 0
        self.coeff = 0
        self.width = 0

        self._cudaRetina = CudaRetina() if gpu else None
        self._fixation = 0 #YX tuple
        self._imsize = 0
        self._gaussNorm = 0 #image
        self._gaussNormTight = 0 #image
        self._normFixation = 0 #YX tuple
        self._V = 0
        self._backproj = 0 #image
        self._backprojTight = 0 #image
        
    def info(self):
        print("loc - an Nx7 array containing retinal nodes defined as follows:\n\
    [x, y, d, angle (radians), dist_5, rf_sigma, rf_width]\n\
coeff - an array of variable size gaussian receptive field kernels\n\
V - the imagevector, output of retinal sampling\n\
gaussNorm - Gaussian normalization image for producing backprojections\n\
\n\
REMEMBER: all coordinates are tuples in the Y,X order, not X,Y.\n\
The only exception is the loc array\n\
REMEMBER2: coeff is redundantly wrapped in another matrix for backwards compatibility")
    
    def loadLocNB(self, path):
        print("This function should not be needed anymore! If a file still requires it, contact Piotr.")
        self.loc = loadPickleNonbin(path)
        self.N = len(self.loc)
        self.width = 2*int(np.abs(self.loc[:,:2]).max() + self.loc[:,6].max()/2.0)
        
    def loadLoc(self, input):
        self.loc = input
        self.N = len(self.loc)
        self.width = 2*int(np.abs(self.loc[:,:2]).max() + self.loc[:,6].max()/2.0)
        
    def loadCoeffNB(self, path):
        print("This function should not be needed anymore! If a file still requires it, contact Piotr.")
        self.coeff = loadPickleNonbin(path)
        
    def loadCoeff(self, input):
        self.coeff = input

    def validate(self):
        assert(len(self.loc) == len(self.coeff[0]))
        if self._gaussNormTight is 0: self._normTight()
        
    def _normTight(self): 
        """Produce a tight-fitted Gaussian normalization image (width x width)"""
        GI = np.zeros((self.width, self.width))         
        r = self.width/2.0
        for i in range(self.N - 1, -1, -1): 
            GI = project(self.coeff[0,i], GI, self.loc[i,:2][::-1] + r)
        
        self._gaussNormTight = GI
        
    def prepare(self, shape, fix):
        """Pre-compute fixation specific Gaussian normalization image """
        fix = (int(fix[0]), int(fix[1]))
        self.validate()
        self._normFixation = fix
        
        GI = np.zeros(shape[:2])
        GI = project(self._gaussNormTight, GI, fix)
        self._gaussNorm = GI
        if self._cudaRetina:
            self._cudaRetina.set_samplingfields(self.loc, self.coeff)
    
    def sample(self, image, fix):
        """Sample an image"""
        fix = (int(fix[0]), int(fix[1]))
        # self.validate()
        self._fixation = fix
        # This will reset the image size only when it was changed.
        if self._imsize != image.shape[:2]:
            self._imsize = image.shape
        if self._cudaRetina:
            # TODO: helper function
            self._cudaRetina.image_width = image.shape[1]
            self._cudaRetina.image_height = image.shape[0]
            self._cudaRetina.rgb = len(image.shape) == 3 and image.shape[-1] == 3
            self._cudaRetina.center_x = int(fix[1])
            self._cudaRetina.center_y = int(fix[0])
            self._cudaRetina.set_gauss_norm(self._gaussNorm)
            V = self._cudaRetina.sample(np.uint8(image)) #XXX uint8 added
            self._V = V
            return V

        rgb = len(image.shape) == 3 and image.shape[-1] == 3
        p = self.width
        pic = pad(image, p, True)
        
        X = self.loc[:,0] + fix[1] + p
        Y = self.loc[:,1] + fix[0] + p
        
        if rgb: V = np.zeros((self.N,3))
        else: V = np.zeros((self.N))
        
        for i in range(0,self.N):
            w = self.loc[i,6]
            y1 = int(Y[i] - w/2+0.5)
            y2 = int(Y[i] + w/2+0.5)
            x1 = int(X[i] - w/2+0.5)
            x2 = int(X[i] + w/2+0.5)
            extract = pic[y1:y2,x1:x2]
            
            c = self.coeff[0, i]
            if rgb: kernel = np.dstack((c,c,c))
            else: kernel = c
            
            m = np.where(np.isnan(extract), 0, 1.0) #mask
            
            if rgb: f = 1.0/np.sum(m*kernel, axis = (0,1)) #TODO fix invalid value warnings
            else: f = 1.0/np.sum(m*kernel)
            
            extract = np.nan_to_num(extract)
            if rgb: V[i] = np.sum(extract*kernel, axis=(0,1)) * f
            else: V[i] = np.sum(extract*kernel) * f
       
        self._V = V
        return V
    
    def backproject_last(self, n = True):
        return self.backproject(self._V, self._imsize, self._fixation, normalize=n)
    
    def backproject(self, V, shape, fix, normalize=True):
        """Backproject the image vector onto a blank matrix equal in size to
         the input image"""
        #TODO: Pyramid requires skipping uint8 cast, which is deeply integrated into GPU codes
        fix = (int(fix[0]), int(fix[1]))
        self.validate()
        if fix != self._normFixation or shape[:2] != self._gaussNorm.shape: 
            self.prepare(shape, fix)
            if self._cudaRetina:
                # TODO: helper
                self._cudaRetina.image_width = shape[1]
                self._cudaRetina.image_height = shape[0]
                self._cudaRetina.rgb = len(shape) == 3 and shape[-1] == 3
                self._cudaRetina.center_x = fix[1]
                self._cudaRetina.center_y = fix[0]
        if self._cudaRetina:
            self._cudaRetina.set_gauss_norm(self._gaussNorm)
            return self._cudaRetina.backproject(V, normalize)
        
        rgb = len(shape) == 3 and shape[-1] == 3
        m = shape[:2]
        
        if rgb: I1 = np.zeros((m[0], m[1], 3))
        else: I1 = np.zeros(m)
        #w = self.width
        #I1 = pad(I1, w, False)        
        
        for i in range(self.N-1,-1,-1):    
            c = self.coeff[0, i]
            if rgb: c = np.dstack((c,c,c))
            location = self.loc[i,:2][::-1] + fix
            if (location > (0, 0)).all() and (location < I1.shape).all():
                I1 = project(c*V[i], I1, location)
        
        GI = self._gaussNorm
        if rgb: GI = np.dstack((GI,GI,GI))
        if normalize: I1 = np.uint8(np.true_divide(I1,GI)) 
        
        self._backproj = I1
        return I1
    
    def backproject_tight_last(self, n=True, norm=None):
        return self.backproject_tight(self._V, self._imsize, self._fixation, normalize=n, norm=norm)
    
    def backproject_tight(self, V, shape, fix, normalize=True, norm=None):
        """Produce a tight-fitted backprojection (width x width, lens only)"""
        fix = (int(fix[0]), int(fix[1]))
        #TODO: look at the weird artifacts at edges when the lens is too big for the frame. CPU version
        
        self.validate()
        if fix != self._normFixation or shape[:2] != self._gaussNorm.shape: 
            self.prepare(shape, fix)
            
        rgb = len(shape) == 3 and shape[-1] == 3
        m = self.width
        r = m/2.0    
        
        if self._cudaRetina:
            self._cudaRetina.image_width = self.width
            self._cudaRetina.image_height = self.width
            self._cudaRetina.rgb = rgb
            self._cudaRetina.center_x = self.width//2
            self._cudaRetina.center_y = self.width//2
            if norm == None: self._cudaRetina.set_gauss_norm(self._gaussNormTight)
            else: self._cudaRetina.set_gauss_norm(norm)
            return self._cudaRetina.backproject(V, normalize)
             
        if rgb: I1 = np.zeros((m, m, 3))
        else: I1 = np.zeros((m, m))
        
        for i in range(self.N - 1,-1,-1):
            c = self.coeff[0, i]
            if rgb: c = np.dstack((c,c,c))
            
            I1 = project(c*V[i], I1, self.loc[i,:2][::-1] + r)
    
        GI = self._gaussNormTight
        if rgb: GI = np.dstack((GI,GI,GI)) #TODO: fix invalid value warnings
        if normalize: 
            I1 = np.uint8(np.true_divide(I1,GI)) 
            self._backprojTight = I1
        return I1
    
    #TODO: add the crop function. Tightly crop original image using retinal lens


from Cython.Build import cythonize
from setuptools import setup, Extension
import numpy
import os


extension_modules = [
    Extension(
        "src.software_retina.utils", 
        ["src/software_retina/utils.pyx"],
        include_dirs=[numpy.get_include()],
    ),
    Extension(
        "src.software_retina.retina",
        ["src/software_retina/retina.pyx"],
        extra_compile_args=["-fopenmp" ],
        extra_link_args=['-fopenmp'],
        include_dirs=[numpy.get_include()],
    ),
    Extension(
        "src.software_retina.rf_generation",
        ["src/software_retina/rf_generation.pyx"],
        include_dirs=[numpy.get_include()],
    )
]

setup(
    name='software-retina',
    version='0.0.14',
    author='Han M. Loo',
    author_email='nloo33755@gmail.com',
    packages=['src', 'src.software_retina', 'src.software_retina_generation', 'src.retinavision'],
    package_data={'src.software_retina': ['*.pyx', '*.pxd']},
    url='https://github.com/hanl00/software-retina',
    description='A software retina inspired by the biological vision system',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    ext_modules=cythonize(extension_modules, force=True), #remove force recompile in final version
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Cython",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    zip_safe=False,
)

# python setup.py build_ext --inplace


# cython: boundscheck=False
# cython: cdivision=True
# cython: wraparound=False

import numpy as np
cimport numpy as cnp

cpdef cnp.int32_t[:, ::1] pad_grayscaled(cnp.ndarray[cnp.uint8_t, ndim=2] img, int padding)

cpdef cnp.int32_t[:, :, ::1] pad_coloured(cnp.ndarray[cnp.uint8_t, ndim=3] img, int padding)

cdef double multiply_and_sum2d(cnp.int32_t[:, ::1] image_extract, cnp.int32_t[:, ::1] coeff_mem_view) nogil

cdef cnp.float64_t[::1] multiply_and_sum3d(cnp.int32_t[:, :, ::1] image_extract, cnp.int32_t[:, ::1] coeff_mem_view, cnp.float64_t[::1] sum3d_return) nogil

cpdef cnp.float64_t gauss(cnp.float64_t sigma, cnp.float64_t x, cnp.float64_t y, int mean)

cpdef cnp.ndarray[cnp.float64_t, ndim=2] gausskernel(cnp.int_t width, cnp.ndarray[cnp.float64_t, ndim=1] loc, cnp.float64_t sigma)



import numpy as np

from .ssnn import SelfSimilarNeuralNetwork
from .utils import normalize, cartesian_to_polar
from .utils import polar_to_cartesian, randomize, point_generation

# Original code provided by George Killick


def fibonacci_sunflower(node_count):

    golden_ratio = (np.sqrt(5) + 1) / 2
    nodes = np.arange(1, node_count+1)
    rho = np.sqrt(nodes-0.5)/np.sqrt(node_count)
    theta = np.pi*2*golden_ratio*nodes
    x = rho*np.cos(theta)
    y = rho*np.sin(theta)

    return np.array([x, y]).T


def fibonacci_retina(node_count, foveal_region_size, foveal_density):

    x = fibonacci_sunflower(node_count)
    x = normalize(x)
    x = cartesian_to_polar(x)
    x[:, 1] *= (1/(foveal_region_size + ((2*np.pi*foveal_region_size) /
                foveal_density)) ** x[:, 1] ** foveal_density)
    x = polar_to_cartesian(x)

    return normalize(x)



# -*- coding: utf-8 -*-
"""
Created on Thu Jul 06 11:33:45 2017

A Python version for Sumitha's receptive field generation code from the file
wilsonretina5fine_8192s.m

@author: Piotr Ozimek
"""
import numpy as np
from scipy.spatial import distance

#Gauss(sigma,x,y) function, 1D
def gauss(sigma,x,y,mean=0):
    d = np.linalg.norm(np.array([x,y]))
    return np.exp(-(d-mean)**2/(2*sigma**2))/np.sqrt(2*np.pi*sigma**2)

#Kernel(width,loc,sigma,x,y)
def gausskernel(width,loc,sigma):
    w = float(width)
    #location is passed as np array [x,y]
    k = np.zeros((width, width))    
    
    shift = (w-1)/2.0

    #subpixel accurate coords of gaussian centre
    dx = loc[0] - int(loc[0])
    dy = loc[1] - int(loc[1])    

#    m = np.sqrt((np.round(loc[0],decimals=1) - int(loc[0]))**2 + (np.round(loc[1],decimals=1) - int(loc[1]))**2)
    
    for x in range(width):
        for y in range(width):
            k[y,x] = gauss(sigma,(x-shift)-dx,(y-shift)-dy)
    
    return k    

def tess_scale(tessellation, neighbourhood, min_rf):
    #compute node density metric dist_5
    d = distance.cdist(tessellation, tessellation)
    s = np.sort(d)
    dist_5 = np.mean(s[:,1:neighbourhood], 1)                   #r
    
    #compute dist_5 for most central 5 nodes
    fov_dist_5 = np.mean(dist_5[:5])                            #closest_rfs
    
    #set central_dist_5 to min_rf (impose min_rf parameter)
    scaled = tessellation*(1/fov_dist_5)*min_rf
    
    return scaled

#Returns offsets (iirc)
def xy_sumitha(x,y,k_width):
    k_width = int(k_width) #this will change                 #d
    
    #if odd size mask -> round coordinates
    if k_width%2 != 0:
        cx = round(x)
        cy = round(y)
        
    #else if even size mask -> 1 decimal point coordinates (always .5)
    else:
        cx = round(x) + np.sign(x-round(x))*0.5
        cy = round(y) + np.sign(y-round(y))*0.5
    
    return cx, cy


"""A python version for Sumithas receptive field generation function. Output
should be nearly identical to the matlab version """
def rf_sumitha(tessellation, min_rf, sigma_ratio, sigma):
    rf_loc = np.zeros((len(tessellation),7))
    rf_coeff = np.ndarray((1,len(tessellation)),dtype='object')
    
    neighbourhood = 6 #5 closest nodes
    
    #compute node density metric dist_5
    d = distance.cdist(tessellation, tessellation)
    s = np.sort(d)
    dist_5 = np.mean(s[:,1:neighbourhood], 1)                   #r
    
    #set central_dist_5 to min_rf (impose min_rf parameter)
    fov_dist_5 = np.mean(dist_5[:5])                            #closest_rfs
    
    
    scaled = tessellation*(1/fov_dist_5)*min_rf
    rf_loc[:,:2] = scaled
    
    rf_loc[:,6] = np.ceil(sigma*(1/sigma_ratio)*(1/fov_dist_5)*min_rf*dist_5)
    
    for i in range(len(tessellation)):
        k_width = int(rf_loc[i,6])
        cx, cy = xy_sumitha(rf_loc[i,0], rf_loc[i,1], k_width)        
        
        rx = rf_loc[i][0] - cx
        ry = rf_loc[i][1] - cy
        loc = np.array([rx, ry])        
        
        #set rf_loc[i,0] (x) rf_loc[i,1] (y) appropriately.
        rf_loc[i,0] = cx
        rf_loc[i,1] = cy
        
        #place proper gaussian in rf_coeff[i]
        rf_coeff[0,i] = gausskernel(k_width,loc, k_width*sigma_ratio)
        rf_coeff[0,i] /= np.sum(rf_coeff[0,i])
        #but why does sumitha divied the gaussian by its sum? Ah to have a 'valid'
        #clipped gaussian that sums to 1?
          
    return rf_loc, rf_coeff

""" Completed function (I hope)
New rf gen function. New features + striving to maintain backwards compatibility
Params:
@tessellation is the raw node locations [x,y] array
@kernel_ratio is the ratio of kernel to local node density (dist_5)
@sigma_base is the base sigma, or global sigma scaling factor (lambda)
@sigma_power is the power term applied to sigma scaling with eccentricity
@mean_rf sets the mean distance between the 20 most central nodes
@min_kernel imposes a minimum kernel width for the receptive fields (def: 3)

Returns:
@rf_loc - a num_of_nodes x 7 array that describes each node as follows:
[x, y, d, angle (radians), dist_5, rf_sigma, rf_width]
@rf_coeff - an array of gaussian receptive field kernels (variable size)
"""
def rf_ozimek(tessellation, kernel_ratio, sigma_base, sigma_power, mean_rf, min_kernel=3):
    rf_loc = np.zeros((len(tessellation), 7))
    rf_coeff = np.ndarray((1, len(tessellation)),dtype='object')
    
    neighbourhood = 6 #5 closest nodes
    
    print("rf generation - might take a while...")
    
    #compute node density metric dist_5
    ##Break up cdist matrix into smaller pieces so they fit in ram
    length = len(tessellation)
    chunk = 5000
    num = length/chunk
    if length%chunk != 0:
        num += 1
    
    dist_5 = np.zeros(length, dtype='float64')
    print(str(chunk) + " nodes in one chunk.")
    for i in range(int(num)):
        print("Processing chunk " + str(i))
        d = distance.cdist(tessellation[i*chunk:(i+1)*chunk], tessellation)
        s = np.sort(d)
        dist_5[i*chunk:(i+1)*chunk] = np.mean(s[:,1:neighbourhood], 1)
    
    #compute min dist_5 for most central 20 nodes
    fov_dist_5 = np.min(dist_5[:20])
    
    #set fov_dist_5 to mean_rf (impose mean_rf parameter) @TODO: this is MIN not MEAN, fix headings
    rf_loc[:,:2] = tessellation*(1/fov_dist_5)*mean_rf
    
    #Adjust dist_5 to reflect new scale
    dist_5 = dist_5*(1/fov_dist_5)*mean_rf
    
    #Insert angle to [3]
    rf_loc[:,3] = np.arctan2(rf_loc[:,1],rf_loc[:,0])
    
    #Insert dist_5 to [4]
    rf_loc[:,4] = dist_5

    print("All chunks done.")
    
    ##determine sigmas
    #sigma_power decreases sigma if dist_5 < 1.0, so correct it
#    correction = 1-mean_rf if mean_rf < 1 else 0
    correction = 0
    rf_loc[:,5] = sigma_base * (dist_5+correction)**sigma_power
    
    # print(rf_loc[:,:2])
    # print(rf_loc[:,:3])
    # print(rf_loc[:,:4])
    # print(rf_loc[:,:5])
    
    #Use the same method as Sumitha for having even/odd kernels [compatibility]
    for i in range(len(tessellation)):
        #determine and insert kernel widths
        k_width = max(min_kernel, int(np.ceil(kernel_ratio*rf_loc[i,4])))
        rf_loc[i,6] = k_width
        
        cx, cy = xy_sumitha(rf_loc[i,0], rf_loc[i,1], k_width)
        
        #Obtain subpixel accurate offsets from kernel centre
        rx = rf_loc[i][0] - cx
        ry = rf_loc[i][1] - cy
        loc = np.array([rx, ry])
        
        #set [2] = eccentricity
        rf_loc[i,2] = np.linalg.norm(rf_loc[i,:2])
        
        #Set x, y, d
        rf_loc[i,0] = cx
        rf_loc[i,1] = cy
        
        #place proper gaussian in rf_coeff[i]
        rf_coeff[0,i] = gausskernel(k_width, loc, rf_loc[i,5])
        rf_coeff[0,i] /= np.sum(rf_coeff[0,i]) ###NORMALIZATION
    
    return rf_loc, rf_coeff, fov_dist_5


"""
function [Ind,M,closest_rfs]=wilsonretina5fine_8192s(retina,min_rf,sigma_ratio,s);
% scale for large RFs for gabors
% example sigma_ratio = 1/6

% min_rf_location=0.03; % on the 0.01th node of the sorted retina EDITED OUT 14.07.2005 
neighbourhood=6; % number of neighbours for density calculation [PO: set to 6]
%retina=retina*0.8;

for i=1:size(retina,1),
    D=distance(retina(i,:)',retina');
    d=sort(D);
    r(i)=mean(d(2:neighbourhood));  % average of the closest, because Paul said use density [PO: mean neighbourhood distance]
end

% EDITED OUT 14.07.2005
% r(1:round(size(r,2)*min_rf_location))=r(round(size(r,2)*min_rf_location));  % ignore smallest rfs in fovea. take minimum from rest
% closest_rfs=min(r(1:size(r,2))); 
closest_rfs=mean(r(1:5)); 

Ind(:,1:2)=retina*(1/closest_rfs)*min_rf; % scale retina so minimum rf gap is 1 pixel NEW " *min_rf "

Ind(:,7)=ceil(s*(1/sigma_ratio)*(1/closest_rfs)*min_rf*r'); % twice average rf gap in neighbourhood - scaled by ()
% (1/closest_rfs) is there because the rf sizes need to be scaled so min_rf

% EDITED OUT 14.07.2005
% Ind(1:min(find((Ind(:,7)==6)))-1,7)=5; % minimum rf size is 5

%%%%%%%%%%%%%%%%%%%%%%%%%%%% Masks %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[S,ignore]=size(Ind);

for i=1:S,
    
    % if i==1000, keyboard; end
    % size of mask
    d=Ind(i,7);  %MARKED LINE POZ
    
    if d/2-fix(d/2)>0, %odd size mask
        cx=round(Ind(i,1)); cy=round(Ind(i,2));
%         Ind(i,4)=cx-(d/2)+0.5; 
%         Ind(i,5)=cy-(d/2)+0.5; 
        X=Ind(i,1)-cx; Y=Ind(i,2)-cy;
        
    else % even size mask
        cx=round(Ind(i,1))+sign(Ind(i,1)-round(Ind(i,1)))*0.5;
        cy=round(Ind(i,2))+sign(Ind(i,2)-round(Ind(i,2)))*0.5;    
        %Ind(i,4)=cx-(d/2)+0.5;
        %Ind(i,5)=cy-(d/2)+0.5;
        X=Ind(i,1)-cx; Y=Ind(i,2)-cy;
    end
    
    Ind(i,1)=cx; Ind(i,2)=cy;
    % M{i}=mask(d,d/(6*1.6),-X,-Y,'dog',d/6);
    
    M{i}=mask(d,d*sigma_ratio,-X,-Y,'gau',d*sigma_ratio);
    M{i} = M{i}/sum(sum(M{i}));
    
end
"""
###################
"""
function M=mask(siz,sigma,X,Y,type,sigma2);
% siz can be even or odd
% X,Y denotes the subpixel distance of the centre of the
% gaussian from the centre of the mask
% M=mask(siz,sigma,X,Y,type,sigma2);

M=zeros(siz);

[x,y] = meshgrid(-(siz-1)/2:(siz-1)/2,-(siz-1)/2:(siz-1)/2);
x=x+X;
y=y+Y;

if type=='gau',
    h = (1/(2*pi*sigma2^2))*exp(-(x.*x + y.*y)/(2*sigma2*sigma2));
"""